import inspect
import logging
import os
from importlib import import_module

from apistar import App, ASyncApp, Include

from .printer import Printer

logger = logging.getLogger('autoapp')


def find_caller_dir():
    """
    This does a little work beyond relying on the caller always
    being at a certain point in the call stack.
    The algorithm is as such:
        * prefer the first module who's filename is app.py
        * otherwise is no app.py is found, use the first module that is
          not this module
    """
    closest_frame = None
    this_file = os.path.basename(__file__)

    for frame in inspect.stack():
        if os.path.basename(frame.filename) == 'app.py':
            return os.path.dirname(frame.filename)

        if not closest_frame and os.path.basename(frame.filename) != this_file:
            closest_frame = frame

    # Didn't find an `app.py`, return the first thing not this.
    if closest_frame:
        return os.path.dirname(closest_frame.filename)

    raise ImportError('Unable to find calling module')


def list_apps(app_dir: str) -> list:
    """
    For a given directory, return all the 'app's in it.
    """
    apps = []

    for de in os.scandir(app_dir):
        if de.is_dir() and os.path.isfile(os.path.join(app_dir, de.name, 'app.py')):
            apps.append(de.name)

    return apps


def inspect_app(app_path: tuple) -> dict:
    """
    Return the attributes need for the App() instance from this app's app.py
    """
    path_str = '.'.join(app_path + ('app',))
    app_mod = import_module(path_str)

    return {
        'routes': getattr(app_mod, 'routes', []),
        'components': getattr(app_mod, 'components', []),
        'template_dir': getattr(app_mod, 'template_dir', []),
        'static_dir': getattr(app_mod, 'static_dir', ''),
        'packages': getattr(app_mod, 'packages', []),
        'event_hooks': getattr(app_mod, 'event_hooks', []),
        'app_path': app_path,
    }


def process_app_routes(app):
    """
    For a given dict of app data figure out if the app is nested
    and if so convert the routes to an Include with the appropriate
    URL prefix.
    """

    routes = app.get('routes')
    app_path = app.get('app_path')

    # If the app has no routes or is not nested (no or empty app_path), do nothing.
    if not routes or not app_path:
        return app

    # Build an Include for the set of nested URLs with the matching prefix
    # for the directory structure
    app['routes'] = [Include(
        '/' + '/'.join(app_path),
        name=':'.join(app_path),
        routes=routes,
    )]

    return app


def find_apps(cur_dir: str, parent_app: tuple = None):
    res = []

    sub_apps = list_apps(cur_dir)
    parent_app = parent_app if parent_app else tuple()

    for app_name in sub_apps:
        app_path = parent_app + (app_name,)
        cur_app = inspect_app(app_path)

        # Decend into each found app
        sub_res = find_apps(os.path.join(cur_dir, app_name), app_path)
        res += [cur_app] + sub_res

    return res


def prioritize(priority_apps, reg_apps):
    if not priority_apps:
        return reg_apps

    # Sort the reg_apps so the priority_apps are honored and first
    papps = []
    apps = reg_apps.copy()

    for papp in priority_apps:
        papp_path = tuple(papp.split('.'))

        for sapp in apps:
            if papp_path == sapp.get('app_path', tuple()):
                papps.append(sapp)
                apps.remove(sapp)

    return papps + apps


def app_args(project_dir: str = None,
             priority_apps: list = None,
             print_results: bool = False,
             **kwargs) -> dict:

    sub_routes = []
    sub_comps = []
    sub_hooks = []

    project_dir = project_dir or find_caller_dir()

    # Find all the sub apps and collect their info then prioritize them
    sub_apps = prioritize(
        priority_apps,
        find_apps(os.path.realpath(project_dir))
    )

    # Collect and process all the sub app data
    for sapp in sub_apps:
        sapp = process_app_routes(sapp)
        sub_routes += sapp.get('routes', [])
        sub_comps += sapp.get('components', [])
        sub_hooks += sapp.get('event_hooks', [])

    # Append the sub app data into the existing app data giving precedence to the
    # existing data
    kwargs['components'] = kwargs.get('components', []) + sub_comps
    kwargs['event_hooks'] = kwargs.get('event_hooks', []) + sub_hooks
    kwargs['routes'] = kwargs.get('routes', []) + sub_routes

    if print_results:
        pr = Printer()
        pr.components(kwargs['components'])
        pr.routes(kwargs['routes'])
        print('')

    return kwargs


def AutoApp(**kwargs) -> App:
    return App(**app_args(**kwargs))


def AutoASyncApp(**kwargs) -> ASyncApp:
    return ASyncApp(**app_args(**kwargs))



import inspect

_tab = '  '


def _sig(func) -> str:
    argspec = inspect.getfullargspec(func)

    # Build the return type string if it's annotated
    return_str = ''
    return_spec = argspec.annotations.get('return')
    if return_spec:
        return_str = ' -> {}:'.format(return_spec.__name__)

    # Build a dict of each args default value
    #     Pad the values of the defaults tuple to match the args list,
    #     then zip the two together to create a dict for easy lookup later
    def_vals = argspec.defaults or tuple()
    def_vals = (('',) * (len(argspec.args) - len(def_vals))) + def_vals
    arg_defs = dict(zip(argspec.args, def_vals))

    # Build a signature string from args, annotations and defaults
    args_str = ''
    for arg in argspec.args:
        spec = argspec.annotations.get(arg, '')

        spec_str = ''
        if spec:
            spec_str = ': {}'.format(spec.__name__)

        adef = arg_defs.get(arg)
        if adef:
            spec_str = '{}={}'.format(spec_str, adef)

        args_str += '{}{}, '.format(arg, spec_str)

    args_str = args_str[:-2]

    return (args_str, return_str)


class Printer(object):
    def __init__(self, printer=None):
        self._pr = printer or print

    def include(self, include, level: int = 0):
        self._pr('{}Include: {} {}'.format(_tab * level, include.url, include.name))

        for route in include.routes:
            self.route(route, parent_url=include.url, level=level + 1)

    def route(self, route, parent_url: str = None, level: int = 0):
        url = parent_url + route.url if parent_url else route.url

        args_str, return_str = _sig(route.handler)

        route_handler = route.handler.__module__ + '.' + route.handler.__qualname__
        route_handler = '{}.{}({}){}'.format(
            route.handler.__module__, route.handler.__qualname__, args_str, return_str
        )

        self._pr('{}Route: {:>3} {}, {}'.format(_tab * level, route.method, url, route.name))
        self._pr('{} {}'.format(_tab * (level + 3), route_handler))

    def routes(self, routes):
        self._pr('\nRoutes:')

        for route in routes:
            if hasattr(route, 'routes'):
                self.include(route, level=1)
            else:
                self.route(route, level=1)

    def components(self, components):
        self._pr('\nComponents:')
        for comp in components:
            args_str, return_str = _sig(comp.resolve)
            self._pr('\t{}'.format(comp.__class__.__name__))
            self._pr('\t\tresolve({}){}'.format(args_str, return_str))



from apistar_autoapp.autoapp import AutoApp, AutoASyncApp, app_args
from apistar_autoapp.printer import Printer

__version__ = '0.5.2'


__all__ = [
    'app_args', 'AutoApp', 'AutoASyncApp', 'Printer'
]
